@@ -28,8 +28,7 @@
 
 #include <locale.h>
 
-static void
-test_basic(void)
+DEFINE_TEST(test_read_format_rar_basic)
 {
   char buff[64];
   const char reffile[] = "test_read_format_rar.rar";
@@ -42,7 +41,7 @@ test_basic(void)
   assert((a = archive_read_new()) != NULL);
   assertA(0 == archive_read_support_filter_all(a));
   assertA(0 == archive_read_support_format_all(a));
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -102,8 +101,7 @@ test_basic(void)
   assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 }
 
-static void
-test_subblock(void)
+DEFINE_TEST(test_read_format_rar_subblock)
 {
   char buff[64];
   const char reffile[] = "test_read_format_rar_subblock.rar";
@@ -116,7 +114,7 @@ test_subblock(void)
   assert((a = archive_read_new()) != NULL);
   assertA(0 == archive_read_support_filter_all(a));
   assertA(0 == archive_read_support_format_all(a));
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -136,8 +134,7 @@ test_subblock(void)
   assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 }
 
-static void
-test_noeof(void)
+DEFINE_TEST(test_read_format_rar_noeof)
 {
   char buff[64];
   const char reffile[] = "test_read_format_rar_noeof.rar";
@@ -150,7 +147,7 @@ test_noeof(void)
   assert((a = archive_read_new()) != NULL);
   assertA(0 == archive_read_support_filter_all(a));
   assertA(0 == archive_read_support_format_all(a));
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -170,8 +167,7 @@ test_noeof(void)
   assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 }
 
-static void
-test_unicode_UTF8(void)
+DEFINE_TEST(test_read_format_rar_unicode_UTF8)
 {
   char buff[30];
   const char reffile[] = "test_read_format_rar_unicode.rar";
@@ -188,7 +184,7 @@ test_unicode_UTF8(void)
   assert((a = archive_read_new()) != NULL);
   assertA(0 == archive_read_support_filter_all(a));
   assertA(0 == archive_read_support_format_all(a));
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -296,8 +292,7 @@ test_unicode_UTF8(void)
   assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 }
 
-static void
-test_unicode_CP932(void)
+DEFINE_TEST(test_read_format_rar_unicode_CP932)
 {
   char buff[30];
   const char reffile[] = "test_read_format_rar_unicode.rar";
@@ -322,7 +317,7 @@ test_unicode_CP932(void)
 	assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 	return;
   }
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -392,8 +387,7 @@ test_unicode_CP932(void)
   assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 }
 
-static void
-test_compress_normal(void)
+DEFINE_TEST(test_read_format_rar_compress_normal)
 {
   const char reffile[] = "test_read_format_rar_compress_normal.rar";
   char file1_buff[20111];
@@ -412,7 +406,7 @@ test_compress_normal(void)
   assert((a = archive_read_new()) != NULL);
   assertA(0 == archive_read_support_filter_all(a));
   assertA(0 == archive_read_support_format_all(a));
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -490,8 +484,7 @@ test_compress_normal(void)
 /* This test is for sufficiently large files that would have been compressed
  * using multiple lzss blocks.
  */
-static void
-test_multi_lzss_blocks(void)
+DEFINE_TEST(test_read_format_rar_multi_lzss_blocks)
 {
   const char reffile[] = "test_read_format_rar_multi_lzss_blocks.rar";
   const char test_txt[] = "-bottom: 0in\"><BR>\n</P>\n</BODY>\n</HTML>";
@@ -504,7 +497,7 @@ test_multi_lzss_blocks(void)
   assert((a = archive_read_new()) != NULL);
   assertA(0 == archive_read_support_filter_all(a));
   assertA(0 == archive_read_support_format_all(a));
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -529,8 +522,7 @@ test_multi_lzss_blocks(void)
   assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 }
 
-static void
-test_compress_best(void)
+DEFINE_TEST(test_read_format_rar_compress_best)
 {
   const char reffile[] = "test_read_format_rar_compress_best.rar";
   char file1_buff[20111];
@@ -549,7 +541,7 @@ test_compress_best(void)
   assert((a = archive_read_new()) != NULL);
   assertA(0 == archive_read_support_filter_all(a));
   assertA(0 == archive_read_support_format_all(a));
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -627,8 +619,7 @@ test_compress_best(void)
 /* This is a test for RAR files compressed using a technique where compression
  * switches back and forth to and from ppmd and lzss decoding.
  */
-static void
-test_ppmd_lzss_conversion(void)
+DEFINE_TEST(test_read_format_rar_ppmd_lzss_conversion)
 {
   const char reffile[] = "test_read_format_rar_ppmd_lzss_conversion.rar";
   const char test_txt[] = "gin-bottom: 0in\"><BR>\n</P>\n</BODY>\n</HTML>";
@@ -641,7 +632,7 @@ test_ppmd_lzss_conversion(void)
   assert((a = archive_read_new()) != NULL);
   assertA(0 == archive_read_support_filter_all(a));
   assertA(0 == archive_read_support_format_all(a));
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -667,8 +658,7 @@ test_ppmd_lzss_conversion(void)
   assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 }
 
-static void
-test_binary(void)
+DEFINE_TEST(test_read_format_rar_binary)
 {
   const char reffile[] = "test_read_format_rar_binary_data.rar";
   char file1_buff[1048576];
@@ -692,7 +682,7 @@ test_binary(void)
   assert((a = archive_read_new()) != NULL);
   assertA(0 == archive_read_support_filter_all(a));
   assertA(0 == archive_read_support_format_all(a));
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -725,8 +715,7 @@ test_binary(void)
   assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 }
 
-static void
-test_windows(void)
+DEFINE_TEST(test_read_format_rar_windows)
 {
   char buff[441];
   const char reffile[] = "test_read_format_rar_windows.rar";
@@ -739,7 +728,7 @@ test_windows(void)
   assert((a = archive_read_new()) != NULL);
   assertA(0 == archive_read_support_filter_all(a));
   assertA(0 == archive_read_support_format_all(a));
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -798,8 +787,225 @@ test_windows(void)
   assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 }
 
-static void
-test_sfx(void)
+DEFINE_TEST(test_read_format_rar_multivolume)
+{
+  const char *reffiles[] =
+  {
+    "test_read_format_rar_multivolume.part0001.rar",
+    "test_read_format_rar_multivolume.part0002.rar",
+    "test_read_format_rar_multivolume.part0003.rar",
+    "test_read_format_rar_multivolume.part0004.rar",
+    NULL
+  };
+  int file1_size = 241647978, offset = 0;
+  char buff[64];
+  const char file1_test_txt[] = "gin-bottom: 0in\"><BR>\n</P>\n</BODY>\n"
+                                "</HTML>";
+  char file2_buff[20111];
+  int file2_size = sizeof(file2_buff);
+  const char file2_test_txt[] = "<P STYLE=\"margin-bottom: 0in\"><BR>\n"
+                                "</P>\n"
+                                "</BODY>\n"
+                                "</HTML>";
+  char file3_buff[20];
+  int file3_size = sizeof(file3_buff);
+  const char file3_test_txt[] = "test text document\r\n";
+  struct archive_entry *ae;
+  struct archive *a;
+
+  extract_reference_files(reffiles);
+  assert((a = archive_read_new()) != NULL);
+  assertA(0 == archive_read_support_filter_all(a));
+  assertA(0 == archive_read_support_format_all(a));
+  assertA(0 == archive_read_open_filenames(a, reffiles, 10240));
+
+  /* First header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("ppmd_lzss_conversion_test.txt",
+                    archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file1_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+  while (offset + (int)sizeof(buff) < file1_size)
+  {
+    assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+    offset += sizeof(buff);
+  }
+  assertA(file1_size - offset ==
+    archive_read_data(a, buff, file1_size - offset));
+  assertEqualMem(buff, file1_test_txt, file1_size - offset);
+
+  /* Second header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("LibarchiveAddingTest.html", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file2_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+  assertA(file2_size == archive_read_data(a, file2_buff, file2_size));
+  assertEqualMem(&file2_buff[file2_size - sizeof(file2_test_txt) + 1],
+                 file2_test_txt, sizeof(file2_test_txt) - 1);
+
+  /* Third header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("testlink", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(0, archive_entry_size(ae));
+  assertEqualInt(41471, archive_entry_mode(ae));
+  assertEqualString("LibarchiveAddingTest.html", archive_entry_symlink(ae));
+  assertEqualIntA(a, 0, archive_read_data(a, file2_buff, 30));
+
+  /* Fourth header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("testdir/test.txt", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file3_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+  assertA(file3_size == archive_read_data(a, file3_buff, file3_size));
+  assertEqualMem(&file3_buff[file3_size + 1 - sizeof(file3_test_txt)],
+                 file3_test_txt, sizeof(file3_test_txt) - 1);
+
+  /* Fifth header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("testdir/LibarchiveAddingTest.html",
+                    archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file2_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+  assertA(file2_size == archive_read_data(a, file2_buff, file2_size));
+  assertEqualMem(&file2_buff[file2_size - sizeof(file2_test_txt) + 1],
+                 file2_test_txt, sizeof(file2_test_txt) - 1);
+
+  /* Sixth header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("testdir", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(0, archive_entry_size(ae));
+  assertEqualInt(16877, archive_entry_mode(ae));
+
+  /* Seventh header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("testemptydir", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(0, archive_entry_size(ae));
+  assertEqualInt(16877, archive_entry_mode(ae));
+
+  /* Test EOF */
+  assertA(1 == archive_read_next_header(a, &ae));
+  assertEqualInt(7, archive_file_count(a));
+  assertEqualIntA(a, ARCHIVE_OK, archive_read_close(a));
+  assertEqualInt(ARCHIVE_OK, archive_read_free(a));
+}
+
+DEFINE_TEST(test_read_format_rar_multivolume_skip)
+{
+  const char *reffiles[] =
+  {
+    "test_read_format_rar_multivolume.part0001.rar",
+    "test_read_format_rar_multivolume.part0002.rar",
+    "test_read_format_rar_multivolume.part0003.rar",
+    "test_read_format_rar_multivolume.part0004.rar",
+    NULL
+  };
+  int file1_size = 241647978;
+  int file2_size = 20111;
+  int file3_size = 20;
+  struct archive_entry *ae;
+  struct archive *a;
+
+  extract_reference_files(reffiles);
+  assert((a = archive_read_new()) != NULL);
+  assertA(0 == archive_read_support_filter_all(a));
+  assertA(0 == archive_read_support_format_all(a));
+  assertA(0 == archive_read_open_filenames(a, reffiles, 10240));
+
+  /* First header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("ppmd_lzss_conversion_test.txt",
+                    archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file1_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+
+  /* Second header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("LibarchiveAddingTest.html", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file2_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+
+  /* Third header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("testlink", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(0, archive_entry_size(ae));
+  assertEqualInt(41471, archive_entry_mode(ae));
+  assertEqualString("LibarchiveAddingTest.html", archive_entry_symlink(ae));
+
+  /* Fourth header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("testdir/test.txt", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file3_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+
+  /* Fifth header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("testdir/LibarchiveAddingTest.html",
+                    archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file2_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+
+  /* Sixth header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("testdir", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(0, archive_entry_size(ae));
+  assertEqualInt(16877, archive_entry_mode(ae));
+
+  /* Seventh header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("testemptydir", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(0, archive_entry_size(ae));
+  assertEqualInt(16877, archive_entry_mode(ae));
+
+  /* Test EOF */
+  assertA(1 == archive_read_next_header(a, &ae));
+  assertEqualInt(7, archive_file_count(a));
+  assertEqualIntA(a, ARCHIVE_OK, archive_read_close(a));
+  assertEqualInt(ARCHIVE_OK, archive_read_free(a));
+}
+
+DEFINE_TEST(test_read_format_rar_sfx)
 {
   char buff[441];
   const char reffile[] = "test_read_format_rar_sfx.exe";
@@ -812,7 +1018,7 @@ test_sfx(void)
   assert((a = archive_read_new()) != NULL);
   assertA(0 == archive_read_support_filter_all(a));
   assertA(0 == archive_read_support_format_all(a));
-  assertA(0 == archive_read_open_file(a, reffile, 10240));
+  assertA(0 == archive_read_open_filename(a, reffile, 10240));
 
   /* First header. */
   assertA(0 == archive_read_next_header(a, &ae));
@@ -871,18 +1077,273 @@ test_sfx(void)
   assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 }
 
-DEFINE_TEST(test_read_format_rar)
+DEFINE_TEST(test_read_format_rar_multivolume_stored_file)
 {
-  test_basic();
-  test_subblock();
-  test_noeof();
-  test_unicode_UTF8();
-  test_unicode_CP932();
-  test_compress_normal();
-  test_multi_lzss_blocks();
-  test_compress_best();
-  test_ppmd_lzss_conversion();
-  test_binary();
-  test_windows();
-  test_sfx();
+  const char *reffiles[] =
+  {
+    "test_rar_multivolume_single_file.part1.rar",
+    "test_rar_multivolume_single_file.part2.rar",
+    "test_rar_multivolume_single_file.part3.rar",
+    NULL
+  };
+  char file_buff[20111];
+  int file_size = sizeof(file_buff);
+  const char file_test_txt[] = "<P STYLE=\"margin-bottom: 0in\"><BR>\n"
+                                "</P>\n"
+                                "</BODY>\n"
+                                "</HTML>";
+  struct archive_entry *ae;
+  struct archive *a;
+
+  extract_reference_files(reffiles);
+  assert((a = archive_read_new()) != NULL);
+  assertA(0 == archive_read_support_filter_all(a));
+  assertA(0 == archive_read_support_format_all(a));
+  assertA(0 == archive_read_open_filenames(a, reffiles, 10240));
+
+  /* First header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("LibarchiveAddingTest.html", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+  assertA(file_size == archive_read_data(a, file_buff, file_size));
+  assertEqualMem(&file_buff[file_size - sizeof(file_test_txt) + 1],
+                 file_test_txt, sizeof(file_test_txt) - 1);
+
+  /* Test EOF */
+  assertA(1 == archive_read_next_header(a, &ae));
+  assertEqualInt(1, archive_file_count(a));
+  assertEqualIntA(a, ARCHIVE_OK, archive_read_close(a));
+  assertEqualInt(ARCHIVE_OK, archive_read_free(a));
+}
+
+DEFINE_TEST(test_read_format_rar_multivolume_stored_file_skip)
+{
+  const char *reffiles[] =
+  {
+    "test_rar_multivolume_single_file.part1.rar",
+    "test_rar_multivolume_single_file.part2.rar",
+    "test_rar_multivolume_single_file.part3.rar",
+    NULL
+  };
+  int file_size = 20111;
+  struct archive_entry *ae;
+  struct archive *a;
+
+  extract_reference_files(reffiles);
+  assert((a = archive_read_new()) != NULL);
+  assertA(0 == archive_read_support_filter_all(a));
+  assertA(0 == archive_read_support_format_all(a));
+  assertA(0 == archive_read_open_filenames(a, reffiles, 10240));
+
+  /* First header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("LibarchiveAddingTest.html", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+
+  /* Test EOF */
+  assertA(1 == archive_read_next_header(a, &ae));
+  assertEqualInt(1, archive_file_count(a));
+  assertEqualIntA(a, ARCHIVE_OK, archive_read_close(a));
+  assertEqualInt(ARCHIVE_OK, archive_read_free(a));
+}
+
+DEFINE_TEST(test_read_format_rar_multivolume_seek_data)
+{
+  const char *reffiles[] =
+  {
+    "test_rar_multivolume_single_file.part1.rar",
+    "test_rar_multivolume_single_file.part2.rar",
+    "test_rar_multivolume_single_file.part3.rar",
+    NULL
+  };
+  char buff[64];
+  int file_size = 20111;
+  const char file_test_txt1[] = "d. \n</P>\n<P STYLE=\"margin-bottom: 0in\">"
+                                "<BR>\n</P>\n</BODY>\n</HTML>";
+  const char file_test_txt2[] = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4."
+                                "0 Transitional//EN\">\n<";
+  const char file_test_txt3[] = "mplify writing such tests,\ntry to use plat"
+                                "form-independent codin";
+  const char file_test_txt4[] = "lString</TT> in the example above)\ngenerat"
+                                "e detailed log message";
+  const char file_test_txt5[] = "SS=\"western\">make check</TT> will usually"
+                                " run\n\tall of the tests.";
+  struct archive_entry *ae;
+  struct archive *a;
+
+  extract_reference_files(reffiles);
+  assert((a = archive_read_new()) != NULL);
+  assertA(0 == archive_read_support_filter_all(a));
+  assertA(0 == archive_read_support_format_all(a));
+  assertA(0 == archive_read_open_filenames(a, reffiles, 10240));
+
+  /* First header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("LibarchiveAddingTest.html", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+
+  /* Seek to the end minus 64 bytes */
+  assertA(file_size - (int)sizeof(buff) ==
+    archive_seek_data(a, file_size - sizeof(buff), SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt1, sizeof(file_test_txt1) - 1);
+
+  /* Seek back to the beginning */
+  assertA(0 == archive_seek_data(a, 0, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt2, sizeof(file_test_txt2) - 1);
+
+  /* Seek to the middle of the combined data block */
+  assertA(10054 == archive_seek_data(a, 10054, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt3, sizeof(file_test_txt3) - 1);
+
+  /* Seek to 32 bytes before the end of the first data sub-block */
+  assertA(6860 == archive_seek_data(a, 6860, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt4, sizeof(file_test_txt4) - 1);
+
+  /* Seek to 32 bytes before the end of the second data sub-block */
+  assertA(13752 == archive_seek_data(a, 13752, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt5, sizeof(file_test_txt5) - 1);
+
+  /* Test EOF */
+  assertA(1 == archive_read_next_header(a, &ae));
+  assertEqualInt(1, archive_file_count(a));
+  assertEqualIntA(a, ARCHIVE_OK, archive_read_close(a));
+  assertEqualInt(ARCHIVE_OK, archive_read_free(a));
+}
+
+DEFINE_TEST(test_read_format_rar_multivolume_seek_multiple_files)
+{
+  const char *reffiles[] =
+  {
+    "test_rar_multivolume_multiple_files.part1.rar",
+    "test_rar_multivolume_multiple_files.part2.rar",
+    "test_rar_multivolume_multiple_files.part3.rar",
+    "test_rar_multivolume_multiple_files.part4.rar",
+    "test_rar_multivolume_multiple_files.part5.rar",
+    "test_rar_multivolume_multiple_files.part6.rar",
+    NULL
+  };
+  char buff[64];
+  int file_size = 20111;
+  const char file_test_txt1[] = "d. \n</P>\n<P STYLE=\"margin-bottom: 0in\">"
+                                "<BR>\n</P>\n</BODY>\n</HTML>";
+  const char file_test_txt2[] = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4."
+                                "0 Transitional//EN\">\n<";
+  const char file_test_txt3[] = "mplify writing such tests,\ntry to use plat"
+                                "form-independent codin";
+  const char file_test_txt4[] = "\nfailures. \n</P>\n<H1 CLASS=\"western\"><"
+                                "A NAME=\"Life_cycle_of_a_te";
+  const char file_test_txt5[] = "LE=\"margin-bottom: 0in\">DO use runtime te"
+                                "sts for platform\n\tfeatu";
+  const char file_test_txt6[] = "rough test suite is essential\nboth for ver"
+                                "ifying new ports and f";
+  const char file_test_txt7[] = "m: 0in\">Creates a temporary directory\n\tw"
+                                "hose name matches the na";
+  const char file_test_txt8[] = "lt\ninput file and verify the results. Thes"
+                                "e use <TT CLASS=\"weste";
+  struct archive_entry *ae;
+  struct archive *a;
+
+  extract_reference_files(reffiles);
+  assert((a = archive_read_new()) != NULL);
+  assertA(0 == archive_read_support_filter_all(a));
+  assertA(0 == archive_read_support_format_all(a));
+  assertA(0 == archive_read_open_filenames(a, reffiles, 10240));
+
+  /* First header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("LibarchiveAddingTest2.html", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+
+  /* Seek to the end minus 64 bytes */
+  assertA(file_size - (int)sizeof(buff) ==
+    archive_seek_data(a, file_size - sizeof(buff), SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt1, sizeof(file_test_txt1) - 1);
+
+  /* Seek back to the beginning */
+  assertA(0 == archive_seek_data(a, 0, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt2, sizeof(file_test_txt2) - 1);
+
+  /* Seek to the middle of the combined data block */
+  assertA(10054 == archive_seek_data(a, 10054, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt3, sizeof(file_test_txt3) - 1);
+
+  /* Seek to 32 bytes before the end of the first data sub-block */
+  assertA(7027 == archive_seek_data(a, 7027, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt4, sizeof(file_test_txt4) - 1);
+
+  /* Seek to 32 bytes before the end of the second data sub-block */
+  assertA(14086 == archive_seek_data(a, 14086, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt5, sizeof(file_test_txt5) - 1);
+
+  /* Second header. */
+  assertA(0 == archive_read_next_header(a, &ae));
+  assertEqualString("LibarchiveAddingTest.html", archive_entry_pathname(ae));
+  assertA((int)archive_entry_mtime(ae));
+  assertA((int)archive_entry_ctime(ae));
+  assertA((int)archive_entry_atime(ae));
+  assertEqualInt(file_size, archive_entry_size(ae));
+  assertEqualInt(33188, archive_entry_mode(ae));
+
+  /* Seek to the end minus 64 bytes */
+  assertA(file_size - (int)sizeof(buff) ==
+    archive_seek_data(a, file_size - sizeof(buff), SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt1, sizeof(file_test_txt1) - 1);
+
+  /* Seek back to the beginning */
+  assertA(0 == archive_seek_data(a, 0, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt2, sizeof(file_test_txt2) - 1);
+
+  /* Seek to the middle of the combined data block */
+  assertA(10054 == archive_seek_data(a, 10054, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt3, sizeof(file_test_txt3) - 1);
+
+  /* Seek to 32 bytes before the end of the first data sub-block */
+  assertA(969 == archive_seek_data(a, 969, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt6, sizeof(file_test_txt4) - 1);
+
+  /* Seek to 32 bytes before the end of the second data sub-block */
+  assertA(8029 == archive_seek_data(a, 8029, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt7, sizeof(file_test_txt5) - 1);
+
+  /* Seek to 32 bytes before the end of the third data sub-block */
+  assertA(15089 == archive_seek_data(a, 15089, SEEK_SET));
+  assertA(sizeof(buff) == archive_read_data(a, buff, sizeof(buff)));
+  assertEqualMem(buff, file_test_txt8, sizeof(file_test_txt5) - 1);
+
+  /* Test EOF */
+  assertA(1 == archive_read_next_header(a, &ae));
+  assertEqualInt(2, archive_file_count(a));
+  assertEqualIntA(a, ARCHIVE_OK, archive_read_close(a));
+  assertEqualInt(ARCHIVE_OK, archive_read_free(a));
 }
